"""Test seqlogic.sim finish."""

from deltacycle import create_task, finish, now, run, sleep


def log(s: str):
    print(f"{now():04} {s}")


async def ctl():
    log("CTL enter")
    await sleep(100)

    # Force all PING threads to stop immediately
    log("CTL finish")
    finish()


async def ping(name: str, period: int):
    log(f"{name} enter")
    while True:
        await sleep(period)
        log(f"{name} PING")


EXP1 = """\
0000 CTL enter
0000 FOO enter
0000 BAR enter
0000 FIZ enter
0000 BUZ enter
0003 FOO PING
0005 BAR PING
0006 FOO PING
0007 FIZ PING
0009 FOO PING
0010 BAR PING
0011 BUZ PING
0012 FOO PING
0014 FIZ PING
0015 BAR PING
0015 FOO PING
0018 FOO PING
0020 BAR PING
0021 FIZ PING
0021 FOO PING
0022 BUZ PING
0024 FOO PING
0025 BAR PING
0027 FOO PING
0028 FIZ PING
0030 BAR PING
0030 FOO PING
0033 BUZ PING
0033 FOO PING
0035 FIZ PING
0035 BAR PING
0036 FOO PING
0039 FOO PING
0040 BAR PING
0042 FIZ PING
0042 FOO PING
0044 BUZ PING
0045 BAR PING
0045 FOO PING
0048 FOO PING
0049 FIZ PING
0050 BAR PING
0051 FOO PING
0054 FOO PING
0055 BUZ PING
0055 BAR PING
0056 FIZ PING
0057 FOO PING
0060 BAR PING
0060 FOO PING
0063 FIZ PING
0063 FOO PING
0065 BAR PING
0066 BUZ PING
0066 FOO PING
0069 FOO PING
0070 FIZ PING
0070 BAR PING
0072 FOO PING
0075 BAR PING
0075 FOO PING
0077 BUZ PING
0077 FIZ PING
0078 FOO PING
0080 BAR PING
0081 FOO PING
0084 FIZ PING
0084 FOO PING
0085 BAR PING
0087 FOO PING
0088 BUZ PING
0090 BAR PING
0090 FOO PING
0091 FIZ PING
0093 FOO PING
0095 BAR PING
0096 FOO PING
0098 FIZ PING
0099 BUZ PING
0099 FOO PING
0100 CTL finish
"""


def test_finish(capsys):

    async def main():
        create_task(ctl())
        create_task(ping("FOO", 3))
        create_task(ping("BAR", 5))
        create_task(ping("FIZ", 7))
        create_task(ping("BUZ", 11))

    # Subsequent calls to run() have no effect
    run(main())

    assert capsys.readouterr().out == EXP1



"""Test seqlogic.sim.Semaphore class."""

import pytest

from deltacycle import BoundedSemaphore, Semaphore, create_task, now, run, sleep


def log(s: str):
    print(f"{now():04} {s}")


async def use_acquire_release(sem: Semaphore, name: str, t1: int, t2: int):
    log(f"{name} enter")

    await sleep(t1)

    log(f"{name} attempt acquire")
    await sem.acquire()
    log(f"{name} acquired")

    try:
        await sleep(t2)
    finally:
        log(f"{name} release")
        sem.release()

    await sleep(10)
    log(f"{name} exit")


async def use_with(sem: Semaphore, name: str, t1: int, t2: int):
    log(f"{name} enter")

    await sleep(t1)

    log(f"{name} attempt acquire")
    async with sem:
        log(f"{name} acquired")
        await sleep(t2)
    log(f"{name} release")

    await sleep(10)
    log(f"{name} exit")


EXP = """\
0000 0 enter
0000 1 enter
0000 2 enter
0000 3 enter
0000 4 enter
0000 5 enter
0000 6 enter
0000 7 enter
0010 0 attempt acquire
0010 0 acquired
0011 1 attempt acquire
0011 1 acquired
0012 2 attempt acquire
0012 2 acquired
0013 3 attempt acquire
0013 3 acquired
0014 4 attempt acquire
0015 5 attempt acquire
0016 6 attempt acquire
0017 7 attempt acquire
0020 0 release
0020 4 acquired
0021 1 release
0021 5 acquired
0022 2 release
0022 6 acquired
0023 3 release
0023 7 acquired
0030 0 exit
0030 4 release
0031 1 exit
0031 5 release
0032 2 exit
0032 6 release
0033 3 exit
0033 7 release
0040 4 exit
0041 5 exit
0042 6 exit
0043 7 exit
"""


def test_acquire_release(capsys):

    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_acquire_release(sem, f"{i}", i + 10, 10))

    run(main())

    assert capsys.readouterr().out == EXP


def test_async_with(capsys):

    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_with(sem, f"{i}", i + 10, 10))

    run(main())

    assert capsys.readouterr().out == EXP


def test_bounds():
    async def use_unbounded():
        sem = Semaphore(2)

        await sem.acquire()
        await sem.acquire()
        sem.release()
        sem.release()

        # No exception!
        sem.release()
        assert sem._cnt == 3  # pylint: disable = protected-access

    async def use_bounded():
        sem = BoundedSemaphore(2)

        await sem.acquire()
        await sem.acquire()
        sem.release()
        sem.release()

        # Exception!
        sem.release()

    run(use_unbounded())

    with pytest.raises(ValueError):
        run(use_bounded())



"""Test dining philosophers."""

from enum import Enum, auto
from random import randint

from deltacycle import Lock, create_task, now, run, sleep

# Number of philosophers
N = 5


# Philosopher state
class State(Enum):
    INITIAL = auto()
    THINKING = auto()
    HUNGRY = auto()
    EATING = auto()


# Eat [min, max] time
EAT_TICKS = (50, 100)

# Think [min, max] time
THINK_TICKS = (50, 100)

# Simulation time
T = 1000


# Optional mask to filter print output
# If pmask & (1<<i), print philosopher state updates.
_pmask = (1 << N) - 1


# Philosophers and Forks
state = [State.INITIAL for _ in range(N)]
forks = [Lock() for _ in range(N)]


def init(pmask: int = (1 << N) - 1):
    """Initialize all philosophers and forks."""
    global _pmask, state, forks
    _pmask = pmask
    state = [State.INITIAL for _ in range(N)]
    forks = [Lock() for _ in range(N)]


def _update(i: int, ns: State):
    """Update philosopher[i] state."""
    if _pmask & (1 << i):
        print(f"[{now():08}] P{i} {state[i].name:8} => {ns.name:8}")
    state[i] = ns


async def think(i: int):
    """Philosopher[i] thinks for a random amount of time."""
    _update(i, State.THINKING)
    await sleep(randint(*THINK_TICKS))


async def pick_up_forks(i: int):
    """Philosopher[i] is hungry. Pick up left/right forks."""
    _update(i, State.HUNGRY)

    # Wait on forks in (left, right) order
    first, second = i, (i + 1) % N

    while True:
        # Wait until first fork is available
        await forks[first].acquire()

        # If second fork is available, get it.
        if forks[second].try_acquire():
            break

        # Second fork is NOT available:
        # 1. Release the first fork
        forks[first].release()
        # 2. Swap which fork we're waiting on first
        first, second = second, first


async def eat(i: int):
    """Philosopher[i] eats for a random amount of time."""
    _update(i, State.EATING)
    await sleep(randint(*EAT_TICKS))


def put_down_forks(i: int):
    """Philosopher[i] is not hungry. Put down left/right forks."""
    first, second = i, (i + 1) % N
    forks[first].release()
    forks[second].release()


async def philosopher(i: int):
    while True:
        await think(i)
        await pick_up_forks(i)
        await eat(i)
        put_down_forks(i)


async def main():
    for i in range(N):
        create_task(philosopher(i))


def test_dp(capsys):
    """This is a random algorithm, so we're only doing some basic checks."""
    init()
    run(main(), until=T)
    s = capsys.readouterr().out
    for i in range(N):
        # Verify all philosophers did at least one full cycle
        assert f"P{i} INITIAL  => THINKING" in s
        assert f"P{i} THINKING => HUNGRY" in s
        assert f"P{i} HUNGRY   => EATING" in s
        assert f"P{i} EATING   => THINKING" in s



"""Test seqlogic.sim.wait function."""

import pytest

from deltacycle import create_task, now, run, sleep, wait


def log(s: str):
    print(f"{now():04} {s}")


async def c(i: int, t: int):
    log(f"C{i} enter")
    await sleep(t)
    log(f"C{i} exit")


EXP1 = """\
0000 MAIN enter
0000 C1 enter
0000 C2 enter
0000 C3 enter
0005 C1 exit
0005 MAIN wait done
0010 C2 exit
0015 C3 exit
0015 MAIN exit
"""


def test_first(capsys):

    async def main():
        log("MAIN enter")

        t1 = create_task(c(1, 5))
        t2 = create_task(c(2, 10))
        t3 = create_task(c(3, 15))

        done, pend = await wait([t1, t2, t3], return_when="FIRST_COMPLETED")
        assert done == {t1}
        assert pend == {t2, t3}

        log("MAIN wait done")

        await t2
        await t3

        log("MAIN exit")

    run(main())
    assert capsys.readouterr().out == EXP1


EXP3 = """\
0000 MAIN enter
0000 C1 enter
0000 C2 enter
0000 C3 enter
0005 C1 exit
0010 C2 exit
0015 C3 exit
0015 MAIN exit
"""


def test_all(capsys):

    async def main():
        log("MAIN enter")

        t1 = create_task(c(1, 5))
        t2 = create_task(c(2, 10))
        t3 = create_task(c(3, 15))

        done, pend = await wait([t1, t2, t3], return_when="ALL_COMPLETED")
        assert done == {t1, t2, t3}
        assert not pend

        log("MAIN exit")

    run(main())
    assert capsys.readouterr().out == EXP3


def test_error1():
    async def main():
        await wait([], return_when="invalid")

    with pytest.raises(ValueError):
        run(main())


def test_error2():
    async def main():
        await wait([], return_when="FIRST_EXCEPTION")

    with pytest.raises(NotImplementedError):
        run(main())



"""Delta Cycle"""

from ._sim import (
    ALL_COMPLETED,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    Aggregate,
    BoundedSemaphore,
    CancelledError,
    Event,
    EventLoop,
    FinishError,
    InvalidStateError,
    Lock,
    Semaphore,
    Singular,
    State,
    Task,
    TaskGroup,
    TaskState,
    Value,
    changed,
    create_task,
    del_event_loop,
    finish,
    get_event_loop,
    get_running_loop,
    irun,
    new_event_loop,
    now,
    resume,
    run,
    set_event_loop,
    sleep,
    wait,
)

__all__ = [
    # error
    "CancelledError",
    "FinishError",
    "InvalidStateError",
    # state
    "State",
    "Value",
    "Singular",
    "Aggregate",
    # task
    "Task",
    "TaskState",
    "TaskGroup",
    "create_task",
    # event
    "Event",
    # semaphore
    "Semaphore",
    "BoundedSemaphore",
    "Lock",
    # event_loop
    "EventLoop",
    "get_running_loop",
    "get_event_loop",
    "set_event_loop",
    "new_event_loop",
    "del_event_loop",
    "now",
    "run",
    "irun",
    "finish",
    "sleep",
    "changed",
    "resume",
    # wait
    "FIRST_COMPLETED",
    "FIRST_EXCEPTION",
    "ALL_COMPLETED",
    "wait",
]



"""Test seqlogic.sim.Lock class."""

import pytest

from deltacycle import CancelledError, Task, TaskGroup, create_task, now, run, sleep


def log(s: str):
    print(f"{now():04} {s}")


async def basic_c1():
    log("C1 enter")

    await sleep(10)
    log("C1")
    await sleep(10)
    log("C1")
    await sleep(10)
    log("C1")
    await sleep(10)

    log("C1 exit")


async def basic_c2(t: Task):
    log("C2 enter")

    log("C2 suspend")
    await t
    log("C2 resume")

    await t
    await sleep(10)
    await t

    log("C2 exit")


EXP1 = """\
0000 C1 enter
0000 C2 enter
0000 C2 suspend
0010 C1
0020 C1
0030 C1
0040 C1 exit
0040 C2 resume
0050 C2 exit
"""


def test_basic(capsys):

    async def main():
        t1 = create_task(basic_c1())
        create_task(basic_c2(t1))

    run(main())
    assert capsys.readouterr().out == EXP1


async def cancel_c1():
    log("C1 enter")

    try:
        await sleep(1000)
    except CancelledError:
        log("C1 except")
        raise
    finally:
        log("C1 finally")


async def cancel_c2():
    log("C2 enter")

    task = create_task(cancel_c1())

    await sleep(1)

    log("C2 cancels C1")
    task.cancel()

    try:
        await task
    except CancelledError:
        log("C2 except")
    finally:
        log("C2 finally")

    assert task.done()
    assert task.cancelled()

    # Result should re-raise CancelledError
    with pytest.raises(CancelledError):
        task.result()
    # So should exception
    with pytest.raises(CancelledError):
        task.exception()


EXP2 = """\
0000 C2 enter
0000 C1 enter
0001 C2 cancels C1
0001 C1 except
0001 C1 finally
0001 C2 except
0001 C2 finally
"""


def test_cancel(capsys):
    run(cancel_c2())
    assert capsys.readouterr().out == EXP2


async def group_c1():
    log("C1 enter")
    await sleep(5)
    log("C1 exit")
    return 1


async def group_c2():
    log("C2 enter")
    await sleep(10)
    log("C2 exit")
    return 2


async def group_c3():
    log("C3 enter")
    await sleep(15)
    log("C3 exit")
    return 3


EXP3 = """\
0000 C1 enter
0000 C2 enter
0000 C3 enter
0005 C1 exit
0010 C2 exit
0015 C3 exit
0015 MAIN
"""


def test_group(capsys):

    async def main():
        async with TaskGroup() as tg:
            t1 = tg.create_task(group_c1())
            t2 = tg.create_task(group_c2())
            t3 = tg.create_task(group_c3())

        log("MAIN")

        assert t1.result() == 1
        assert t2.result() == 2
        assert t3.result() == 3

    run(main())
    assert capsys.readouterr().out == EXP3



"""Test seqlogic.sim module."""

from collections import defaultdict

import pytest

from deltacycle import (
    Singular,
    State,
    create_task,
    get_running_loop,
    irun,
    now,
    resume,
    run,
    sleep,
)
from deltacycle._sim import INIT_TIME

waves = defaultdict(dict)


def _waves_add(time, var, val):
    waves[time][var] = val


class Bool(Singular):
    """Variable that supports dumping to memory."""

    def __init__(self):
        super().__init__(value=False)
        _waves_add(INIT_TIME, self, self._value)

    def update(self):
        if self.dirty():
            _waves_add(now(), self, self._next_value)
        super().update()

    def is_posedge(self) -> bool:
        return not self._value and self._next_value

    def is_negedge(self) -> bool:
        return self._value and not self._next_value

    def is_edge(self) -> bool:
        return self.is_posedge() or self.is_negedge()

    async def posedge(self) -> State:
        await resume((self, self.is_posedge))

    async def negedge(self) -> State:
        await resume((self, self.is_negedge))

    async def edge(self) -> State:
        await resume((self, self.is_edge))


HELLO_OUT = """\
[2] Hello
[4] World
"""


def test_hello(capsys):
    """Test basic async/await hello world functionality."""

    async def hello():
        await sleep(2)
        print(f"[{now()}] Hello")
        await sleep(2)
        print(f"[{now()}] World")

    # Invalid run limit
    with pytest.raises(TypeError):
        run(ticks="Invalid argument type")

    # Run until no events left
    run(hello())

    assert capsys.readouterr().out == HELLO_OUT


def test_vars_run():
    """Test generic variable functionality."""
    waves.clear()

    clk = Bool()
    a = Bool()
    b = Bool()

    async def p_clk():
        while True:
            await sleep(5)
            clk.next = not clk.value

    async def p_a():
        i = 0
        while True:
            await clk.edge()
            if i % 2 == 0:
                a.next = not a.value
            else:
                a.next = a.value
            i += 1

    async def p_b():
        i = 0
        while True:
            await clk.edge()
            if i % 3 == 0:
                b.next = not b.value
            i += 1

    async def main():
        create_task(p_clk())
        create_task(p_a())
        create_task(p_b())

    # Expected sim output
    exp = {
        -1: {clk: False, a: False, b: False},
        5: {clk: True, a: True, b: True},
        10: {clk: False},
        15: {clk: True, a: False},
        20: {clk: False, b: False},
        25: {clk: True, a: True},
        30: {clk: False},
        35: {clk: True, a: False, b: True},
        40: {clk: False},
        45: {clk: True, a: True},
    }

    # Relative run limit
    run(main(), ticks=25)

    # Absolute run limit
    run(loop=get_running_loop(), until=50)

    assert waves == exp


def test_vars_iter():
    """Test generic variable functionality."""
    waves.clear()

    clk = Bool()
    a = Bool()
    b = Bool()

    async def p_clk():
        while True:
            await sleep(5)
            clk.next = not clk.value

    async def p_a():
        i = 0
        while True:
            await clk.edge()
            if i % 2 == 0:
                a.next = not a.value
            else:
                a.next = a.value
            i += 1

    async def p_b():
        i = 0
        while True:
            await clk.edge()
            if i % 3 == 0:
                b.next = not b.value
            i += 1

    async def main():
        create_task(p_clk())
        create_task(p_a())
        create_task(p_b())

    # Expected sim output
    exp = {
        -1: {clk: False, a: False, b: False},
        5: {clk: True, a: True, b: True},
        10: {clk: False},
        15: {clk: True, a: False},
        20: {clk: False, b: False},
        25: {clk: True, a: True},
        30: {clk: False},
        35: {clk: True, a: False, b: True},
        40: {clk: False},
        45: {clk: True, a: True},
    }

    for _ in irun(main(), until=25):
        pass

    for _ in irun(loop=get_running_loop(), until=50):
        pass

    assert waves == exp



"""Test seqlogic.sim.Event class."""

from deltacycle import Event, create_task, now, run, sleep


def log(s: str):
    print(f"{now():04} {s}")


async def primary(event: Event, name: str):
    log(f"{name} enter")
    await sleep(10)
    log(f"{name} set")
    event.set()
    assert event.is_set()
    await sleep(10)
    log(f"{name} exit")


async def secondary(event: Event, name: str):
    log(f"{name} enter")
    log(f"{name} waiting")
    await event.wait()
    log(f"{name} running")
    await sleep(10)
    log(f"{name} exit")


EXP1 = """\
0000 P1 enter
0000 S1 enter
0000 S1 waiting
0000 S2 enter
0000 S2 waiting
0000 S3 enter
0000 S3 waiting
0010 P1 set
0010 S1 running
0010 S2 running
0010 S3 running
0020 P1 exit
0020 S1 exit
0020 S2 exit
0020 S3 exit
"""


def test_acquire_release(capsys):

    async def main():
        event = Event()
        create_task(primary(event, "P1"))
        create_task(secondary(event, "S1"))
        create_task(secondary(event, "S2"))
        create_task(secondary(event, "S3"))

    run(main())

    assert capsys.readouterr().out == EXP1



"""Test seqlogic.sim.Lock class."""

from deltacycle import Lock, create_task, now, run, sleep


def log(s: str):
    print(f"{now():04} {s}")


async def use_acquire_release(lock: Lock, name: str, t1: int, t2: int):
    log(f"{name} enter")

    await sleep(t1)

    log(f"{name} attempt acquire")
    await lock.acquire()
    log(f"{name} acquired")

    try:
        await sleep(t2)
    finally:
        log(f"{name} release")
        lock.release()

    await sleep(10)
    log(f"{name} exit")


async def use_with(lock: Lock, name: str, t1: int, t2: int):
    log(f"{name} enter")

    await sleep(t1)

    log(f"{name} attempt acquire")
    async with lock:
        log(f"{name} acquired")
        await sleep(t2)
    log(f"{name} release")

    await sleep(10)
    log(f"{name} exit")


EXP = """\
0000 0 enter
0000 1 enter
0000 2 enter
0000 3 enter
0010 0 attempt acquire
0010 0 acquired
0011 1 attempt acquire
0012 2 attempt acquire
0013 3 attempt acquire
0020 0 release
0020 1 acquired
0030 0 exit
0030 1 release
0030 2 acquired
0040 1 exit
0040 2 release
0040 3 acquired
0050 2 exit
0050 3 release
0060 3 exit
"""


def test_acquire_release(capsys):

    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_acquire_release(lock, f"{i}", i + 10, 10))

    run(main())

    assert capsys.readouterr().out == EXP


def test_async_with(capsys):

    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_with(lock, f"{i}", i + 10, 10))

    run(main())

    assert capsys.readouterr().out == EXP



"""Simulation using async/await.

We are intentionally imitating the style of Python's Event Loop API:
https://docs.python.org/3/library/asyncio-eventloop.html

Credit to David Beazley's "Build Your Own Async" tutorial for inspiration:
https://www.youtube.com/watch?v=Y4Gt3Xjd7G8
"""

# pylint: disable=protected-access

from __future__ import annotations

import heapq
from abc import ABC
from collections import defaultdict, deque, namedtuple
from collections.abc import Awaitable, Callable, Coroutine, Generator, Hashable
from enum import IntEnum, auto
from typing import override

INIT_TIME = -1
START_TIME = 0

type Predicate = Callable[[], bool]


class CancelledError(Exception):
    """Task has been cancelled."""


class FinishError(Exception):
    """Force the simulation to stop."""


class InvalidStateError(Exception):
    """Task has an invalid state."""


class State(Awaitable):
    """Model component."""

    def __await__(self) -> Generator[None, State, State]:
        _loop.state_wait(self, self.changed)

        # Suspend
        state = yield
        assert state is self

        # Resume
        return state

    present = property(fget=NotImplemented)

    def changed(self) -> bool:
        raise NotImplementedError()  # pragma: no cover

    def update(self):
        raise NotImplementedError()  # pragma: no cover


class Value(ABC):
    """State value."""

    def _get_value(self):
        raise NotImplementedError()  # pragma: no cover

    value = property(fget=_get_value)

    def _set_next(self, value):
        raise NotImplementedError()  # pragma: no cover

    next = property(fset=_set_next)


class Singular(State, Value):
    """Model state organized as a single unit."""

    def __init__(self, value):
        State.__init__(self)
        self._value = value
        self._next_value = value
        self._changed = False

    # Value
    def _get_value(self):
        return self._value

    value = property(fget=_get_value)

    def _set_next(self, value):
        self._changed = value != self._next_value
        self._next_value = value

        # Notify the event loop
        _loop.state_touch(self)

    next = property(fset=_set_next)

    # State
    def _get_present(self):
        return self._next_value

    present = property(fget=_get_present)

    def changed(self) -> bool:
        return self._changed

    def update(self):
        self._value = self._next_value
        self._changed = False

    def dirty(self) -> bool:
        return self._next_value != self._value


class Aggregate(State):
    """Model state organized as multiple units."""

    def __init__(self, value):
        super().__init__()
        self._values = defaultdict(lambda: value)
        self._next_values = dict()

    # [key] => Value
    def __getitem__(self, key: Hashable) -> AggrValue:
        return AggrValue(self, key)

    def _get_value(self, key: Hashable):
        return self._values[key]

    def _get_next_value(self, key: Hashable):
        try:
            return self._next_values[key]
        except KeyError:
            return self._values[key]

    def _set_next(self, key: Hashable, value):
        if value != self._get_next_value(key):
            self._next_values[key] = value

        # Notify the event loop
        _loop.state_touch(self)

    # State
    def _get_present(self) -> AggrPresent:
        return AggrPresent(self)

    present = property(fget=_get_present)

    def changed(self) -> bool:
        return bool(self._next_values)

    def update(self):
        while self._next_values:
            key, value = self._next_values.popitem()
            self._values[key] = value


class AggrPresent:
    """Wrap Aggregate present value."""

    def __init__(self, aggr: Aggregate):
        self._aggr = aggr

    def __getitem__(self, key: Hashable):
        return self._aggr._get_next_value(key)


class AggrValue(Value):
    """Wrap Aggregate value getter/setter."""

    def __init__(self, aggr: Aggregate, key: Hashable):
        self._aggr = aggr
        self._key = key

    def _get_value(self):
        return self._aggr._get_value(self._key)

    value = property(fget=_get_value)

    def _set_next(self, value):
        self._aggr._set_next(self._key, value)

    next = property(fset=_set_next)


class TaskState(IntEnum):
    """Task State.

    Transitions::

                   +--------------------------+
                   |                          |
                   v                          |
        CREATED -> PENDING -> RUNNING -> WAIT_*
                                      -> CANCELLED
                                      -> EXCEPTED
                                      -> RETURNED
    """

    # Default value after instantiation
    CREATED = auto()

    # Awaiting task/event/semaphore in FIFO order
    WAIT_FIFO = auto()

    # Awaiting state touch
    WAIT_STATE = auto()

    # In the event queue
    PENDING = auto()

    # Current running
    RUNNING = auto()

    # Done: returned a result
    RETURNED = auto()
    # Done: raised an exception
    EXCEPTED = auto()
    # Done: cancelled
    CANCELLED = auto()


class Task(Awaitable):
    """Coroutine wrapper."""

    def __init__(self, coro: Coroutine, region: int = 0):
        self._coro = coro
        self._region = region

        self._state = TaskState.CREATED
        self._parent: Task | State | Event | Semaphore | None = None

        self._result = None

        # Set flag to throw exception
        self._exc_flag = False
        self._exception = None

    def __await__(self) -> Generator[None, None, None]:
        if not self.done():
            _loop.fifo_wait(self)
            # Suspend
            yield
        # Resume
        return

    @property
    def region(self):
        return self._region

    def set_state(self, state: TaskState, parent=None):
        self._state = state
        self._parent = parent

    def run(self, value=None):
        self._state = TaskState.RUNNING
        if self._exc_flag:
            self._exc_flag = False
            self._coro.throw(self._exception)
        else:
            self._coro.send(value)

    def done(self) -> bool:
        return self._state in {
            TaskState.CANCELLED,
            TaskState.EXCEPTED,
            TaskState.RETURNED,
        }

    def cancelled(self) -> bool:
        return self._state == TaskState.CANCELLED

    def set_result(self, result):
        self._result = result

    def result(self):
        if self._state == TaskState.CANCELLED:
            assert self._exception is not None and isinstance(self._exception, CancelledError)
            raise self._exception
        if self._state == TaskState.EXCEPTED:
            assert self._exception is not None
            raise self._exception  # Re-raise exception
        if self._state == TaskState.RETURNED:
            assert self._exception is None
            return self._result
        raise InvalidStateError("Task is not done")

    def set_exception(self, exc):
        self._exc_flag = True
        self._exception = exc

    def exception(self):
        if self._state == TaskState.CANCELLED:
            assert self._exception is not None and isinstance(self._exception, CancelledError)
            raise self._exception
        if self._state == TaskState.EXCEPTED:
            assert self._exception is not None
            return self._exception  # Return exception
        if self._state == TaskState.RETURNED:
            assert self._exception is None
            return self._exception
        raise InvalidStateError("Task is not done")

    def get_coro(self) -> Coroutine:
        return self._coro

    def cancel(self, msg: str | None = None):
        match self._state:
            case TaskState.WAIT_FIFO:
                _loop.fifo_drop(self._parent, self)
            case TaskState.WAIT_STATE:
                _loop.state_drop(self._parent, self)
            case TaskState.PENDING:
                _loop.drop(self)
            case _:
                raise ValueError("Task is not WAITING or PENDING")

        args = () if msg is None else (msg,)
        exc = CancelledError(*args)
        self.set_exception(exc)
        _loop.call_soon(self)


def create_task(coro: Coroutine, region: int = 0) -> Task:
    return _loop.task_create(coro, region)


class TaskGroup:
    """Group of tasks."""

    def __init__(self):
        self._tasks = deque()

    def create_task(self, coro: Coroutine, region: int = 0) -> Task:
        task = _loop.task_create(coro, region)
        self._tasks.append(task)
        return task

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        while self._tasks:
            task = self._tasks.popleft()
            await task


class Event:
    """Notify multiple tasks that some event has happened."""

    def __init__(self):
        self._flag = False

    async def wait(self):
        if not self._flag:
            _loop.fifo_wait(self)
            await _Awaitable()

    def set(self):
        _loop.event_set(self)
        self._flag = True

    def clear(self):
        self._flag = False

    def is_set(self) -> bool:
        return self._flag


class Semaphore:
    """Semaphore to synchronize tasks."""

    def __init__(self, value: int = 1):
        if value < 1:
            raise ValueError(f"Expected value >= 1, got {value}")
        self._value = value
        self._cnt = value

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self.release()

    async def acquire(self):
        assert self._cnt >= 0
        if self._cnt == 0:
            _loop.fifo_wait(self)
            await _Awaitable()
        else:
            self._cnt -= 1

    def try_acquire(self) -> bool:
        assert self._cnt >= 0
        if self._cnt == 0:
            return False
        self._cnt -= 1
        return True

    def release(self):
        assert self._cnt >= 0
        increment = _loop.sem_release(self)
        if increment:
            self._cnt += 1

    def locked(self) -> bool:
        return self._cnt == 0


class BoundedSemaphore(Semaphore):

    @override
    def release(self):
        assert self._cnt >= 0
        increment = _loop.sem_release(self)
        if increment:
            if self._cnt == self._value:
                raise ValueError("Cannot release")
            self._cnt += 1


class Lock(BoundedSemaphore):
    """Mutex lock to synchronize tasks."""

    def __init__(self):
        super().__init__(value=1)


_TaskQueueItem = namedtuple("_TaskQueueItem", ["time", "region", "index", "task", "value"])


class _TaskQueue:
    """Priority queue for ordering task execution."""

    def __init__(self):
        # time, region, index, task, value
        self._items: list[_TaskQueueItem] = []

        # Monotonically increasing integer to break ties in the heapq
        self._index: int = 0

    def __bool__(self) -> bool:
        return bool(self._items)

    def clear(self):
        self._items.clear()
        self._index = 0

    def push(self, time: int, task: Task, value: State | None = None):
        item = _TaskQueueItem(time, task.region, self._index, task, value)
        heapq.heappush(self._items, item)
        self._index += 1

    def peek(self) -> tuple[int, Task, State | None]:
        time, _, _, task, value = self._items[0]
        return (time, task, value)

    def pop(self) -> tuple[int, Task, State | None]:
        time, _, _, task, value = heapq.heappop(self._items)
        return (time, task, value)

    def pop_time(self) -> Generator[tuple[int, Task, State | None], None, None]:
        time, _, _, task, value = heapq.heappop(self._items)
        yield (time, task, value)
        while self._items:
            t, _, _, task, value = self._items[0]
            if t == time:
                heapq.heappop(self._items)
                yield (time, task, value)
            else:
                break

    def drop(self, task: Task):
        for i, item in enumerate(self._items):
            if item.task is task:
                index = i
                break
        else:
            raise ValueError("Task is not scheduled")
        self._items.pop(index)


class _Awaitable(Awaitable):
    """Suspend execution of the current task."""

    def __await__(self):
        # Suspend
        value = yield
        # Resume
        return value


class EventLoop:
    """Simulation event loop."""

    def __init__(self):
        """Initialize simulation."""
        # Simulation time
        self._time: int = INIT_TIME

        # Task queue
        self._queue = _TaskQueue()

        # Currently executing task
        self._task: Task | None = None

        # Awaitable FIFOs
        self._wait_fifos: dict[Task | Event | Semaphore, deque[Task]] = defaultdict(deque)

        # State waiting set
        self._waiting: dict[State, set[Task]] = defaultdict(set)
        self._predicates: dict[State, dict[Task, Predicate]] = defaultdict(dict)
        self._touched: set[State] = set()

    def clear(self):
        """Clear all task collections."""
        self._queue.clear()
        self._wait_fifos.clear()
        self._waiting.clear()
        self._predicates.clear()
        self._touched.clear()

    def restart(self):
        """Restart current simulation."""
        self._time = INIT_TIME
        self._task = None
        self.clear()

    def time(self) -> int:
        return self._time

    def task(self) -> Task | None:
        return self._task

    # Scheduling methods
    def _schedule(self, time: int, task: Task, value):
        task.set_state(TaskState.PENDING)
        self._queue.push(time, task, value)

    def call_soon(self, task: Task, value=None):
        self._schedule(self._time, task, value)

    def call_later(self, delay: int, task: Task, value=None):
        self._schedule(self._time + delay, task, value)

    def call_at(self, when: int, task: Task, value=None):
        self._schedule(when, task, value)

    def drop(self, task: Task):
        self._queue.drop(task)

    def fifo_wait(self, aw: Task | Event | Semaphore):
        task = self._task
        task.set_state(TaskState.WAIT_FIFO, aw)
        self._wait_fifos[aw].append(task)

    def fifo_drop(self, aw: Task | Event | Semaphore, task: Task):
        self._wait_fifos[aw].remove(task)

    # Task await / done callbacks
    def task_create(self, coro: Coroutine, region: int = 0) -> Task:
        # Cannot call task_create before the simulation starts
        assert self._time >= 0
        task = Task(coro, region)
        self.call_soon(task)
        return task

    def _task_return(self, ptask: Task, e: StopIteration):
        waiting = self._wait_fifos[ptask]
        while waiting:
            ctask = waiting.popleft()
            self.call_soon(ctask, value=ptask)
        ptask.set_state(TaskState.RETURNED)
        ptask.set_result(e.value)

    def _task_cancel(self, ptask: Task, e: CancelledError):
        waiting = self._wait_fifos[ptask]
        while waiting:
            ctask = waiting.popleft()
            ctask.set_exception(e)
            self.call_soon(ctask, value=ptask)
        ptask.set_state(TaskState.CANCELLED)

    def _task_except(self, ptask: Task, e: Exception):
        waiting = self._wait_fifos[ptask]
        while waiting:
            ctask = waiting.popleft()
            ctask.set_exception(e)
            self.call_soon(ctask, value=ptask)
        ptask.set_state(TaskState.EXCEPTED)

    # Event wait / set callbacks
    def event_set(self, event: Event):
        waiting = self._wait_fifos[event]
        while waiting:
            task = waiting.popleft()
            self.call_soon(task, value=event)

    # Semaphore acquire / release callbacks
    def sem_release(self, sem: Semaphore) -> bool:
        waiting = self._wait_fifos[sem]
        if waiting:
            task = waiting.popleft()
            self.call_soon(task, value=sem)
            # Do NOT increment semaphore counter
            return False
        # Increment semaphore counter
        return True

    # State suspend / resume callbacks
    def state_wait(self, state: State, predicate: Predicate):
        """Schedule current coroutine after a state update trigger."""
        task = self._task
        task.set_state(TaskState.WAIT_STATE, state)
        self._waiting[state].add(task)
        self._predicates[state][task] = predicate

    def state_drop(self, state: State, task: Task):
        self._waiting[state].remove(task)
        del self._predicates[state][task]

    def state_touch(self, state: State):
        """Schedule coroutines triggered by touching model state."""
        waiting = self._waiting[state]
        predicates = self._predicates[state]
        pending = [task for task in waiting if predicates[task]()]
        for task in pending:
            self.state_drop(state, task)
            self.call_soon(task, value=state)
        # Add state to update set
        self._touched.add(state)

    def _state_update(self):
        while self._touched:
            state = self._touched.pop()
            state.update()

    def _limit(self, ticks: int | None, until: int | None) -> int | None:
        """Determine the run limit."""
        match ticks, until:
            # Run until no tasks left
            case None, None:
                return None
            # Run until an absolute time
            case None, int():
                return until
            # Run until a number of ticks in the future
            case int(), None:
                return max(START_TIME, self._time) + ticks
            case _:
                s = "Expected either ticks or until to be int | None"
                raise TypeError(s)

    def _run_kernel(self, limit: int | None):
        while self._queue:
            # Peek when next event is scheduled
            time, _, _ = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Exit if we hit the run limit
            if limit is not None and time >= limit:
                break

            # Otherwise, advance to new timeslot
            self._time = time

            # Execute time slot
            for _, task, value in self._queue.pop_time():
                self._task = task
                try:
                    task.run(value)
                except StopIteration as e:
                    self._task_return(task, e)
                except CancelledError as e:
                    self._task_cancel(task, e)
                except Exception as e:
                    self._task_except(task, e)
                    raise

            # Update simulation state
            self._state_update()

    def run(self, ticks: int | None = None, until: int | None = None):
        """Run the simulation.

        Until:
        1. We hit the runlimit, OR
        2. There are no tasks left in the queue
        """
        limit = self._limit(ticks, until)

        # Run until either 1) all tasks complete, or 2) finish()
        try:
            self._run_kernel(limit)
        except FinishError:
            self.clear()

    def _iter_kernel(self, limit: int | None) -> Generator[int, None, None]:
        while self._queue:
            # Peek when next event is scheduled
            time, _, _ = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Exit if we hit the run limit
            if limit is not None and time >= limit:
                return

            # Otherwise, advance to new timeslot
            self._time = time

            # Execute time slot
            for _, task, value in self._queue.pop_time():
                self._task = task
                try:
                    task.run(value)
                except StopIteration as e:
                    self._task_return(task, e)
                except CancelledError as e:
                    self._task_cancel(task, e)
                except Exception as e:
                    self._task_except(task, e)
                    raise

            # Update simulation state
            self._state_update()
            yield self._time

    def irun(
        self, ticks: int | None = None, until: int | None = None
    ) -> Generator[int, None, None]:
        """Iterate the simulation.

        Until:
        1. We hit the runlimit, OR
        2. There are no tasks left in the queue
        """
        limit = self._limit(ticks, until)

        try:
            yield from self._iter_kernel(limit)
        except FinishError:
            self.clear()


_loop: EventLoop | None = None


def get_running_loop() -> EventLoop:
    if _loop is None:
        raise RuntimeError("No running loop")
    return _loop


def get_event_loop() -> EventLoop | None:
    """Get the current event loop."""
    return _loop


def set_event_loop(loop: EventLoop):
    """Set the current event loop."""
    global _loop
    _loop = loop


def new_event_loop() -> EventLoop:
    """Create and return a new event loop."""
    return EventLoop()


def del_event_loop():
    """Delete the current event loop."""
    global _loop
    _loop = None


def now() -> int:
    if _loop is None:
        raise RuntimeError("No running loop")
    return _loop.time()


def run(
    coro: Coroutine | None = None,
    region: int = 0,
    loop: EventLoop | None = None,
    ticks: int | None = None,
    until: int | None = None,
):
    """Run a simulation."""
    global _loop

    task = None
    if loop is not None:
        _loop = loop
    else:
        _loop = EventLoop()
        task = Task(coro, region)
        _loop.call_at(START_TIME, task)

    _loop.run(ticks, until)


def irun(
    coro: Coroutine | None = None,
    region: int = 0,
    loop: EventLoop | None = None,
    ticks: int | None = None,
    until: int | None = None,
) -> Generator[int, None, None]:
    """Iterate a simulation."""
    global _loop

    task = None
    if loop is not None:
        _loop = loop
    else:
        _loop = EventLoop()
        task = Task(coro, region)
        _loop.call_at(START_TIME, task)

    yield from _loop.irun(ticks, until)


async def sleep(delay: int):
    """Suspend the task, and wake up after a delay."""
    _loop.call_later(delay, _loop.task())
    await _Awaitable()


async def changed(*states: State) -> State:
    """Resume execution upon state change."""
    for state in states:
        _loop.state_wait(state, state.changed)
    state = await _Awaitable()
    return state


async def resume(*triggers: tuple[State, Predicate]) -> State:
    """Resume execution upon event."""
    for state, predicate in triggers:
        _loop.state_wait(state, predicate)
    state = await _Awaitable()
    return state


FIRST_COMPLETED = "FIRST_COMPLETED"
FIRST_EXCEPTION = "FIRST_EXCEPTION"
ALL_COMPLETED = "ALL_COMPLETED"


async def wait(aws, return_when=ALL_COMPLETED) -> tuple[set[Task], set[Task]]:
    # TODO(cjdrake): Catch exceptions
    if return_when == FIRST_EXCEPTION:
        raise NotImplementedError("FIRST_EXCEPTION not implemented yet")

    if return_when not in {FIRST_COMPLETED, ALL_COMPLETED}:
        exp = ", ".join([FIRST_COMPLETED, FIRST_EXCEPTION, ALL_COMPLETED])
        s = f"Expected return_when in {{{exp}}}, got {return_when}"
        raise ValueError(s)

    done = set()
    pend = set(aws)

    for aw in aws:
        _loop.fifo_wait(aw)

    while True:
        aw = await _Awaitable()

        done.add(aw)
        pend.remove(aw)

        if return_when == FIRST_COMPLETED and len(done) == 1:
            for aw in pend:
                _loop.fifo_drop(aw, _loop.task())
            break

        if return_when == ALL_COMPLETED and not pend:
            break

    return done, pend


def finish():
    raise FinishError()
